Package: griddy
Title: Geospatial Distribution Dynamics for Simple Features
Version: 0.1.0
Authors@R: 
    person("Dmitry", "Shkolnik", email = "shkolnikd@gmail.com", role = c("aut", "cre"))
Description: Tools for exploratory geospatial distribution dynamics with
    'sf' objects and tidy data. Provides pooled and time-specific
    classification of longitudinal spatial values, classic discrete Markov
    transition matrices, spatial Markov matrices conditioned on spatial-lag
    classes, endpoint and adjacent rank mobility summaries, and 'ggplot2'
    visualizations. Methods follow Rey (2001)
    <doi:10.1111/j.1538-4632.2001.tb00444.x> and Rey et al. (2016)
    <doi:10.1007/s10109-016-0234-x>; design is inspired by the Python
    'PySAL' 'giddy' package <https://pysal.org/giddy/>.
URL: https://github.com/dshkol/griddy, https://dshkol.github.io/griddy/
BugReports: https://github.com/dshkol/griddy/issues
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.2
Depends: R (>= 4.2)
Imports: dplyr, ggplot2, rlang, scales, sf, spdep, tibble, tidyr
Suggests: cancensus, knitr, microbenchmark, pkgdown, rmarkdown, sfdep,
        spData, testthat (>= 3.0.0), tidycensus
VignetteBuilder: knitr
Config/testthat/edition: 3
LazyData: true
NeedsCompilation: no
Packaged: 2026-05-27 16:36:40 UTC; dmitryshkolnik
Author: Dmitry Shkolnik [aut, cre]
Maintainer: Dmitry Shkolnik <shkolnikd@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-30 13:30:02 UTC
Built: R 4.5.2; ; 2026-05-30 17:24:55 UTC; unix

Package: griddy
Title: Geospatial Distribution Dynamics for Simple Features
Version: 0.1.0
Authors@R: 
    person("Dmitry", "Shkolnik", email = "shkolnikd@gmail.com", role = c("aut", "cre"))
Description: Tools for exploratory geospatial distribution dynamics with
    'sf' objects and tidy data. Provides pooled and time-specific
    classification of longitudinal spatial values, classic discrete Markov
    transition matrices, spatial Markov matrices conditioned on spatial-lag
    classes, endpoint and adjacent rank mobility summaries, and 'ggplot2'
    visualizations. Methods follow Rey (2001)
    <doi:10.1111/j.1538-4632.2001.tb00444.x> and Rey et al. (2016)
    <doi:10.1007/s10109-016-0234-x>; design is inspired by the Python
    'PySAL' 'giddy' package <https://pysal.org/giddy/>.
URL: https://github.com/dshkol/griddy, https://dshkol.github.io/griddy/
BugReports: https://github.com/dshkol/griddy/issues
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.2
Depends: R (>= 4.2)
Imports: dplyr, ggplot2, rlang, scales, sf, spdep, tibble, tidyr
Suggests: cancensus, knitr, microbenchmark, pkgdown, rmarkdown, sfdep,
        spData, testthat (>= 3.0.0), tidycensus
VignetteBuilder: knitr
Config/testthat/edition: 3
LazyData: true
NeedsCompilation: no
Packaged: 2026-05-27 16:36:40 UTC; dmitryshkolnik
Author: Dmitry Shkolnik [aut, cre]
Maintainer: Dmitry Shkolnik <shkolnikd@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-30 13:30:02 UTC
Built: R 4.6.0; ; 2026-05-30 18:03:03 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,grd_classes)
S3method(as.data.frame,grd_markov)
S3method(as.data.frame,grd_rank_mobility)
S3method(as.data.frame,grd_spatial_markov)
S3method(class_intervals,grd_classes)
S3method(class_intervals,grd_markov)
S3method(class_intervals,grd_spatial_markov)
S3method(lag_intervals,grd_spatial_markov)
S3method(print,grd_classes)
S3method(print,grd_markov)
S3method(print,grd_rank_mobility)
S3method(print,grd_spatial_markov)
S3method(steady_state,grd_markov)
S3method(steady_state,matrix)
S3method(transition_matrix,grd_markov)
S3method(transition_matrix,grd_spatial_markov)
export(class_intervals)
export(classify_dynamics)
export(lag_intervals)
export(markov_dynamics)
export(plot_rank_mobility)
export(plot_spatial_markov)
export(plot_transition_matrix)
export(rank_mobility)
export(spatial_markov)
export(steady_state)
export(transition_matrix)
importFrom(rlang,.data)

# griddy

`griddy` is an R package for geospatial distribution dynamics with `sf` and tidy data. It is inspired by PySAL `giddy` and by established spatial distribution dynamics work, especially Rey (2001). Prior R implementations exist, including `estdaR` and `spdyn`, but neither is on CRAN or designed along tidy principles. This package is designed to work best with long data and `sf` native spatial workflows.

Current scope:

- classify longitudinal spatial values into comparable classes
- estimate classic Markov transition matrices
- estimate spatial Markov transition matrices conditioned on spatial lag class
- compute simple rank mobility
- return tidy tables and `ggplot2` plots

The initial CRAN release focuses on the core classification, Markov, spatial Markov, rank mobility, and plotting workflows.

## Installation

```r
# install.packages("pak")
pak::pak("dshkol/griddy")
```

The minimal example below also uses `spData` for state geometry:

```r
install.packages("spData")
```

## What it does

`griddy` keeps the workflow keyed by explicit `id`, `time`, and `value`
columns instead of matrix row position. The analytical outputs preserve
transition labels, class intervals, and spatial-lag intervals so they can be
inspected, joined, and plotted without reverse-engineering array dimensions.

## Minimal example

The bundled `usjoin` panel is the canonical PySAL `giddy` reference dataset:
48 contiguous US states, per-capita personal income, 1929 to 2009.

```r
library(griddy)
library(dplyr)
library(sf)
library(sfdep)
library(spData)

data(usjoin)

geom <- us_states |>
  filter(NAME %in% usjoin$name) |>
  arrange(NAME) |>
  mutate(
    nb = st_contiguity(geometry),
    wt = st_weights(nb)
  )

panel <- usjoin |>
  filter(name %in% geom$NAME) |>
  arrange(name, year)

classes <- classify_dynamics(panel, name, year, income, k = 5)
classic <- markov_dynamics(classes, name, year, class)
spatial <- spatial_markov(panel, name, year, income, geometry = geom, k = 5)

classic$transitions |> select(id, from_time, to_time, transition) |> head()
lag_intervals(spatial)
```

`spatial_markov()` takes a `geometry` argument: an `sf` tibble with one row
per spatial unit and `nb` / `wt` list-columns produced by `sfdep`. This keeps
the spatial frame, neighbor structure, and row-standardization choice in one
tidy object. `listw` and `nb` arguments remain accepted for compatibility with
existing workflows.

## Documentation

The pkgdown site is organized around:

- core workflow and concepts
- `tidycensus` and `cancensus` examples showing applications with US and Canadian census data
- prior-art comparison against `estdaR` and `spdyn`
- performance benchmarking notes

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----message=FALSE, warning=FALSE---------------------------------------------
library(griddy)
library(dplyr)
library(ggplot2)
library(sf)
library(sfdep)
library(spData)
library(tidyr)

## ----map-theme, include=FALSE-------------------------------------------------
map_theme <- function() {
  theme_void(base_size = 11) +
    theme(
      legend.position = "bottom",
      plot.title.position = "plot",
      strip.text = element_text(face = "bold")
    )
}

## -----------------------------------------------------------------------------
data(usjoin)

usjoin

## -----------------------------------------------------------------------------
classes <- classify_dynamics(usjoin, name, year, income, k = 5)

class_intervals(classes)

## -----------------------------------------------------------------------------
states <- us_states |>
  st_drop_geometry() |>
  filter(NAME %in% usjoin$name) |>
  pull(NAME)

geom <- us_states |>
  filter(NAME %in% usjoin$name) |>
  arrange(NAME)

## ----fig.alt="Map of contiguous US states showing pooled income classes in 1955 using the same five class intervals estimated from the full 1929 to 2009 panel.", fig.width=6.8, fig.height=4----
class_map <- classes |>
  filter(year == 1955) |>
  left_join(geom |> select(NAME, geometry), by = c("name" = "NAME")) |>
  st_as_sf()

ggplot(class_map) +
  geom_sf(aes(fill = class), color = "white", linewidth = 0.15) +
  scale_fill_viridis_d(option = "C", direction = -1, name = "Income class") +
  coord_sf(datum = NA) +
  labs(
    title = "Pooled classes remain map-ready",
    subtitle = "1955 uses breaks estimated from the full 1929-2009 panel"
  ) +
  map_theme()

## ----fig.alt="Heatmap of classic Markov transition probabilities for US state per-capita income, 1929 to 2009."----
classic <- markov_dynamics(classes, name, year, class)

classic
transition_matrix(classic, "count")
steady_state(classic)

plot_transition_matrix(classic)

## -----------------------------------------------------------------------------
geom <- us_states |>
  filter(NAME %in% usjoin$name) |>
  arrange(NAME) |>
  mutate(
    nb = st_contiguity(geometry),
    wt = st_weights(nb)
  )

panel <- usjoin |>
  filter(name %in% states) |>
  arrange(name, year)

## ----fig.alt="Faceted heatmaps of spatial Markov transition probabilities for US state per-capita income, by spatial-lag quintile."----
spatial <- spatial_markov(panel, name, year, income, geometry = geom, k = 5)

lag_intervals(spatial)
transition_matrix(spatial, "probability", lag_class = "Q1")

plot_spatial_markov(spatial)

## ----fig.alt="Map of contiguous US states showing each state's 1994 spatial-lag income class based on neighboring states.", fig.width=6.8, fig.height=4----
lag_map <- spatial$transitions |>
  filter(from_time == 1994) |>
  distinct(id, lag_class, spatial_lag) |>
  left_join(geom |> select(NAME, geometry), by = c("id" = "NAME")) |>
  st_as_sf()

ggplot(lag_map) +
  geom_sf(aes(fill = lag_class), color = "white", linewidth = 0.15) +
  scale_fill_viridis_d(option = "C", direction = -1, name = "Spatial-lag class") +
  coord_sf(datum = NA) +
  labs(
    title = "Spatial Markov conditions on neighboring-state income",
    subtitle = "Spatial-lag class in 1994"
  ) +
  map_theme()

## ----fig.alt="Map of endpoint rank mobility for US states, 1929 to 2009, with positive values indicating upward rank movement."----
mobility_panel <- usjoin |>
  filter(name %in% states) |>
  left_join(geom |> select(NAME), by = c("name" = "NAME")) |>
  st_as_sf()

mobility <- rank_mobility(mobility_panel, name, year, income)

mobility |>
  st_drop_geometry() |>
  arrange(desc(abs_rank_change)) |>
  select(name, start_rank, end_rank, rank_change) |>
  head()

plot_rank_mobility(mobility) +
  coord_sf(datum = NA) +
  labs(
    title = "Endpoint rank mobility, 1929 to 2009",
    subtitle = "Positive values indicate upward movement in the state income ranking"
  ) +
  map_theme()

## -----------------------------------------------------------------------------
adjacent_mobility <- rank_mobility(panel, name, year, income, compare = "adjacent")

adjacent_mobility |>
  arrange(desc(abs_rank_change)) |>
  select(name, year, to_time, rank_change) |>
  head()

## -----------------------------------------------------------------------------
classic$transitions |> head()
spatial$transitions |> select(id, from_time, to_time, lag_class, transition, spatial_lag) |> head()

#' Classify Longitudinal Values For Distribution Dynamics
#'
#' Converts numeric values in a long panel into ordered classes suitable for
#' transition analysis.
#'
#' @param data A data frame or `sf` object in long format.
#' @param id,time,value Columns identifying spatial unit, time, and value.
#' @param method Classification method. `"pooled_quantile"` uses one set of
#'   quantile breaks across all periods. `"time_quantile"` uses period-specific
#'   ranks. `"fixed"` uses user-supplied `breaks`. `"existing"` treats `value`
#'   as an already classified state.
#' @param k Number of quantile classes.
#' @param breaks Breaks for `method = "fixed"`.
#' @param labels Optional class labels.
#'
#' @return A data frame or `sf` object with a `class` column and class
#'   `grd_classes`.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(letters[1:4], each = 3),
#'   year = rep(2020:2022, times = 4),
#'   value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
#' )
#'
#' classes <- classify_dynamics(panel, id, year, value, k = 3)
#' classes
#' class_intervals(classes)
#' @export
classify_dynamics <- function(data, id, time, value,
                              method = c("pooled_quantile", "time_quantile", "fixed", "existing"),
                              k = 5,
                              breaks = NULL,
                              labels = NULL) {
  method <- match.arg(method)
  id <- rlang::ensym(id)
  time <- rlang::ensym(time)
  value <- rlang::ensym(value)

  value_vec <- dplyr::pull(data, !!value)

  out <- switch(
    method,
    pooled_quantile = {
      q_breaks <- .grd_quantile_breaks(value_vec, k)
      data |>
        dplyr::mutate(class = .grd_cut(!!value, q_breaks, labels = labels))
    },
    time_quantile = {
      data |>
        dplyr::group_by(!!time) |>
        dplyr::mutate(class = .grd_ntile_factor(!!value, k, labels = labels)) |>
        dplyr::ungroup()
    },
    fixed = {
      if (is.null(breaks)) {
        stop("`breaks` is required when `method = \"fixed\"`.", call. = FALSE)
      }
      data |>
        dplyr::mutate(class = .grd_cut(!!value, breaks, labels = labels))
    },
    existing = {
      state_labels <- labels %||% sort(unique(stats::na.omit(as.character(value_vec))))
      data |>
        dplyr::mutate(class = factor(as.character(!!value), levels = state_labels, ordered = TRUE))
    }
  )

  out <- out |>
    dplyr::arrange(!!id, !!time)

  class(out) <- unique(c("grd_classes", class(out)))
  attr(out, "id") <- rlang::as_name(id)
  attr(out, "time") <- rlang::as_name(time)
  attr(out, "value") <- rlang::as_name(value)
  attr(out, "method") <- method
  attr(out, "k") <- k
  attr(out, "breaks") <- if (method == "pooled_quantile") .grd_quantile_breaks(value_vec, k) else breaks
  attr(out, "class_intervals") <- if (method %in% c("pooled_quantile", "fixed")) {
    .grd_break_table(attr(out, "breaks"), labels = labels, type = "value")
  } else {
    NULL
  }
  out
}

#' @export
print.grd_classes <- function(x, ...) {
  cat("<grd_classes>\n")
  NextMethod()
}

#' @export
as.data.frame.grd_classes <- function(x, ...) {
  class(x) <- setdiff(class(x), "grd_classes")
  as.data.frame(x, ...)
}

test_that("all-upward panel produces a strictly upper-triangular count matrix", {
  dat <- tibble::tibble(
    id = rep(letters[1:3], each = 3),
    year = rep(1:3, times = 3),
    state = factor(
      c("low", "mid", "high", "low", "mid", "high", "low", "mid", "high"),
      levels = c("low", "mid", "high"),
      ordered = TRUE
    )
  )

  mk <- markov_dynamics(dat, id, year, state)

  counts <- unname(mk$counts)
  expect_equal(counts[lower.tri(counts, diag = TRUE)], rep(0, 6))
  expect_equal(counts[1, 2], 3)
  expect_equal(counts[2, 3], 3)
})

test_that("reducible chain with an absorbing state has an absorbing row", {
  dat <- tibble::tibble(
    id = rep(c("trap", "free"), each = 4),
    year = rep(1:4, times = 2),
    state = factor(
      c("a", "a", "a", "a", "a", "b", "b", "b"),
      levels = c("a", "b"),
      ordered = TRUE
    )
  )

  mk <- markov_dynamics(dat, id, year, state)

  expect_equal(unname(mk$probabilities)[2, ], c(0, 1))
})

test_that("row probabilities sum to 1 on non-empty rows and are NA on empty rows", {
  dat <- tibble::tibble(
    id = rep(c("a", "b"), each = 3),
    year = rep(1:3, times = 2),
    state = factor(
      c("low", "low", "high", "low", "high", "high"),
      levels = c("low", "high"),
      ordered = TRUE
    )
  )

  mk <- markov_dynamics(dat, id, year, state)
  probs <- mk$probabilities

  for (i in seq_len(nrow(probs))) {
    row_total <- sum(mk$counts[i, ])
    if (row_total > 0) {
      expect_equal(sum(probs[i, ]), 1, tolerance = 1e-12)
    } else {
      expect_true(all(is.na(probs[i, ])))
    }
  }
})

test_that("spatial Markov on a tiny known grid agrees with hand calculation", {
  panel <- tibble::tibble(
    id = rep(1:4, each = 2),
    year = rep(c(1, 2), times = 4),
    value = c(
      1, 1,
      1, 2,
      2, 2,
      2, 3
    )
  )

  listw <- spdep::nb2listw(spdep::cell2nb(2, 2, type = "queen"), style = "W")
  spatial <- spatial_markov(panel, id, year, value, listw = listw, k = 2)

  total_observed <- sum(spatial$matrix$n)
  expect_equal(total_observed, 4)

  for (lag in spatial$lag_states) {
    sub <- dplyr::filter(spatial$matrix, .data$lag_class == lag)
    by_from <- sub |>
      dplyr::group_by(.data$from_state) |>
      dplyr::summarise(prob_total = sum(.data$probability), n_total = sum(.data$n), .groups = "drop")
    for (i in seq_len(nrow(by_from))) {
      if (by_from$n_total[i] > 0) {
        expect_equal(by_from$prob_total[i], 1, tolerance = 1e-12)
      } else {
        expect_true(is.na(by_from$prob_total[i]) || by_from$prob_total[i] == 0)
      }
    }
  }
})

test_that("results are stable under row reordering when id and time are supplied", {
  dat <- tibble::tibble(
    id = rep(c("a", "b", "c"), each = 3),
    year = rep(2000:2002, times = 3),
    value = c(1, 2, 3, 4, 5, 6, 7, 8, 9)
  )
  shuffled <- dat[sample.int(nrow(dat)), ]

  mk1 <- markov_dynamics(classify_dynamics(dat, id, year, value, k = 3), id, year, class)
  mk2 <- markov_dynamics(classify_dynamics(shuffled, id, year, value, k = 3), id, year, class)

  expect_equal(unname(mk1$counts), unname(mk2$counts))
  expect_equal(unname(mk1$probabilities), unname(mk2$probabilities))
})

test_that("rank mobility endpoint changes sum to zero on a complete untied panel", {
  dat <- tibble::tibble(
    id = rep(letters[1:5], each = 2),
    year = rep(c(1, 2), times = 5),
    value = c(10, 14, 20, 21, 30, 32, 40, 41, 50, 13)
  )

  mob <- rank_mobility(dat, id, year, value)
  expect_equal(sum(mob$rank_change), 0)
})

#' Plot A Transition Matrix
#'
#' @param x A `grd_markov` object.
#'
#' @return A `ggplot` object.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(id = rep(1:4, each = 2), year = rep(2020:2021, 4), value = 1:8)
#' markov <- panel |>
#'   classify_dynamics(id, year, value, k = 2) |>
#'   markov_dynamics(id, year, class)
#'
#' plot_transition_matrix(markov)
#' @export
plot_transition_matrix <- function(x) {
  stopifnot(inherits(x, "grd_markov"))
  x$matrix |>
    ggplot2::ggplot(ggplot2::aes(.data$to_state, .data$from_state, fill = .data$probability)) +
    ggplot2::geom_tile(color = "white", linewidth = 0.35) +
    ggplot2::geom_text(ggplot2::aes(label = scales::percent(.data$probability, accuracy = 1)), na.rm = TRUE, size = 3) +
    ggplot2::scale_fill_viridis_c(labels = scales::percent, na.value = "grey90") +
    ggplot2::labs(x = "To state", y = "From state", fill = "Probability")
}

#' Plot Spatial Markov Matrices
#'
#' @param x A `grd_spatial_markov` object.
#'
#' @return A `ggplot` object.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(1:4, each = 2),
#'   year = rep(2020:2021, times = 4),
#'   value = c(1, 2, 2, 3, 4, 3, 5, 6)
#' )
#' grid <- sf::st_sf(
#'   id = 1:4,
#'   geometry = sf::st_make_grid(
#'     sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2)),
#'     n = c(2, 2)
#'   )
#' ) |>
#'   dplyr::mutate(
#'     nb = sfdep::st_contiguity(geometry),
#'     wt = sfdep::st_weights(nb)
#'   )
#' spatial <- spatial_markov(panel, id, year, value, geometry = grid, k = 2)
#' plot_spatial_markov(spatial)
#' @export
plot_spatial_markov <- function(x) {
  stopifnot(inherits(x, "grd_spatial_markov"))
  x$matrix |>
    ggplot2::ggplot(ggplot2::aes(.data$to_state, .data$from_state, fill = .data$probability)) +
    ggplot2::geom_tile(color = "white", linewidth = 0.3) +
    ggplot2::facet_wrap(ggplot2::vars(.data$lag_class)) +
    ggplot2::scale_fill_viridis_c(labels = scales::percent, na.value = "grey90") +
    ggplot2::labs(x = "To state", y = "From state", fill = "Probability")
}

#' Plot Rank Mobility
#'
#' @param x A `grd_rank_mobility` object returned by [rank_mobility()].
#'
#' @return A `ggplot` object.
#' @export
plot_rank_mobility <- function(x) {
  if (!inherits(x, "sf")) {
    stop("`x` must retain geometry to draw a rank mobility map.", call. = FALSE)
  }
  ggplot2::ggplot(x) +
    ggplot2::geom_sf(ggplot2::aes(fill = .data$rank_change), color = "white", linewidth = 0.2) +
    ggplot2::scale_fill_gradient2(
      low = "#b2182b",
      mid = "grey95",
      high = "#2166ac",
      midpoint = 0
    ) +
    ggplot2::labs(fill = "Rank gain")
}

make_grid_panel <- function() {
  geom <- sf::st_as_sf(
    tibble::tibble(
      id = c("a", "b", "c", "d"),
      x = c(0, 1, 0, 1),
      y = c(0, 0, 1, 1)
    ),
    coords = c("x", "y")
  ) |>
    sf::st_buffer(0.45)

  panel <- geom[rep(seq_len(nrow(geom)), each = 2), ] |>
    dplyr::mutate(
      year = rep(c(1, 2), times = 4),
      value = c(1, 3, 2, 4, 3, 2, 4, 1)
    ) |>
    dplyr::arrange(id, year)

  list(geom = geom, panel = panel)
}

test_that("spatial_markov conditional counts aggregate to classic counts", {
  fx <- make_grid_panel()
  nb <- spdep::cell2nb(2, 2, type = "queen")

  classes <- classify_dynamics(fx$panel, id, year, value, k = 2)
  mk <- markov_dynamics(classes, id, year, class)
  sm <- spatial_markov(fx$panel, id, year, value, nb = nb, k = 2)

  conditional_counts <- sm$matrix |>
    dplyr::group_by(from_state, to_state) |>
    dplyr::summarise(n = sum(n), .groups = "drop") |>
    dplyr::arrange(from_state, to_state)

  classic_counts <- as.data.frame(mk$counts) |>
    tibble::as_tibble() |>
    dplyr::transmute(from_state = Var1, to_state = Var2, n = Freq) |>
    dplyr::arrange(from_state, to_state)

  expect_equal(conditional_counts$n, classic_counts$n)
  expect_equal(nrow(lag_intervals(sm)), 2)
  expect_true(all(c("lag_lower", "lag_upper", "transition") %in% names(sm$matrix)))
  expect_true("transition" %in% names(sm$transitions))
})

test_that("rank_mobility endpoint changes sum to zero for complete untied ranks", {
  fx <- make_grid_panel()

  rm <- rank_mobility(fx$panel, id, year, value)

  expect_s3_class(rm, "grd_rank_mobility")
  expect_equal(sum(sf::st_drop_geometry(rm)$rank_change), 0)
  expect_true("geometry" %in% names(rm))
})

test_that("plot helpers return ggplot objects", {
  fx <- make_grid_panel()
  nb <- spdep::cell2nb(2, 2, type = "queen")
  classes <- classify_dynamics(fx$panel, id, year, value, k = 2)
  mk <- markov_dynamics(classes, id, year, class)
  sm <- spatial_markov(fx$panel, id, year, value, nb = nb, k = 2)
  rm <- rank_mobility(fx$panel, id, year, value)

  expect_s3_class(plot_transition_matrix(mk), "ggplot")
  expect_s3_class(plot_spatial_markov(sm), "ggplot")
  expect_s3_class(plot_rank_mobility(rm), "ggplot")
})

#' Estimate Classic Markov Transition Dynamics
#'
#' Estimates transition counts and probabilities between adjacent periods for a
#' classified long panel.
#'
#' @param classes A classified data frame from [classify_dynamics()] or any data
#'   frame with ID, time, and state columns.
#' @param id,time,state Columns identifying spatial unit, time, and state.
#'
#' @return A `grd_markov` object.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(letters[1:4], each = 3),
#'   year = rep(2020:2022, times = 4),
#'   value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
#' )
#'
#' classes <- classify_dynamics(panel, id, year, value, k = 3)
#' markov <- markov_dynamics(classes, id, year, class)
#'
#' markov
#' transition_matrix(markov)
#' steady_state(markov)
#' @export
markov_dynamics <- function(classes, id, time, state = class) {
  id <- rlang::ensym(id)
  time <- rlang::ensym(time)
  state <- rlang::ensym(state)

  data <- .grd_drop_geometry(classes)
  state_vec <- dplyr::pull(data, !!state)
  state_levels <- if (is.factor(state_vec)) levels(state_vec) else sort(unique(stats::na.omit(as.character(state_vec))))

  transitions <- data |>
    dplyr::arrange(!!id, !!time) |>
    dplyr::group_by(!!id) |>
    dplyr::mutate(
      to_state = dplyr::lead(!!state),
      to_time = dplyr::lead(!!time)
    ) |>
    dplyr::ungroup() |>
    dplyr::filter(!is.na(.data$to_state)) |>
    dplyr::transmute(
      id = !!id,
      from_time = !!time,
      to_time = .data$to_time,
      from_state = factor(as.character(!!state), levels = state_levels, ordered = TRUE),
      to_state = factor(as.character(.data$to_state), levels = state_levels, ordered = TRUE)
    ) |>
    dplyr::mutate(
      transition = paste(.data$from_state, "->", .data$to_state),
      .after = "to_state"
    )

  counts <- table(transitions$from_state, transitions$to_state)
  probabilities <- .grd_prob_table(counts)

  matrix_data <- .grd_complete_matrix_tibble(transitions, state_levels)

  out <- list(
    transitions = transitions,
    matrix = matrix_data,
    counts = counts,
    probabilities = probabilities,
    states = state_levels,
    class_intervals = attr(classes, "class_intervals"),
    id = rlang::as_name(id),
    time = rlang::as_name(time),
    state = rlang::as_name(state)
  )
  class(out) <- "grd_markov"
  out
}

#' Extract A Transition Matrix
#'
#' @param x A `grd_markov` or `grd_spatial_markov` object.
#' @param type `"probability"` or `"count"`.
#' @param lag_class Optional lag class for `grd_spatial_markov`.
#'
#' @return A matrix.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(letters[1:4], each = 3),
#'   year = rep(2020:2022, times = 4),
#'   value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
#' )
#' markov <- panel |>
#'   classify_dynamics(id, year, value, k = 3) |>
#'   markov_dynamics(id, year, class)
#'
#' transition_matrix(markov, "count")
#' @export
transition_matrix <- function(x, type = c("probability", "count"), lag_class = NULL) {
  type <- match.arg(type)
  UseMethod("transition_matrix")
}

#' @export
transition_matrix.grd_markov <- function(x, type = c("probability", "count"), lag_class = NULL) {
  type <- match.arg(type)
  if (type == "count") x$counts else x$probabilities
}

#' @export
transition_matrix.grd_spatial_markov <- function(x, type = c("probability", "count"), lag_class = NULL) {
  type <- match.arg(type)
  if (is.null(lag_class)) {
    stop("`lag_class` is required for spatial Markov objects.", call. = FALSE)
  }
  target_lag_class <- as.character(lag_class)
  dat <- x$matrix |>
    dplyr::filter(as.character(.data$lag_class) == .env$target_lag_class)
  if (nrow(dat) == 0) {
    stop("No transition matrix found for `lag_class`.", call. = FALSE)
  }
  value <- if (type == "count") "n" else "probability"
  wide <- dat |>
    dplyr::select("from_state", "to_state", value = dplyr::all_of(value)) |>
    dplyr::group_by(.data$from_state, .data$to_state) |>
    dplyr::summarise(value = dplyr::first(.data$value), .groups = "drop") |>
    tidyr::pivot_wider(names_from = "to_state", values_from = "value") |>
    tibble::column_to_rownames("from_state")
  as.matrix(wide)
}

#' Estimate Stationary Distribution
#'
#' @param x A `grd_markov` object or transition probability matrix.
#'
#' @return A numeric vector.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' prob <- matrix(c(0.8, 0.2, 0.3, 0.7), nrow = 2, byrow = TRUE)
#' steady_state(prob)
#' @export
steady_state <- function(x) {
  UseMethod("steady_state")
}

#' @export
steady_state.grd_markov <- function(x) {
  out <- .grd_stationary(x$probabilities)
  names(out) <- x$states
  out
}

#' @export
steady_state.matrix <- function(x) {
  out <- .grd_stationary(x)
  names(out) <- rownames(x)
  out
}

#' @export
print.grd_markov <- function(x, ...) {
  cat("<grd_markov>\n")
  cat(length(x$states), "states,", nrow(x$transitions), "observed transitions\n")
  print(x$probabilities)
  invisible(x)
}

#' @export
as.data.frame.grd_markov <- function(x, ...) {
  as.data.frame(x$matrix, ...)
}

#' Class Intervals Used By A griddy Object
#'
#' @param x A griddy result object.
#' @param ... Reserved for future methods.
#'
#' @return A tibble of class intervals when available.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(id = rep(1:3, each = 2), year = rep(2020:2021, 3), value = 1:6)
#' classes <- classify_dynamics(panel, id, year, value, k = 3)
#' class_intervals(classes)
#' @export
class_intervals <- function(x, ...) {
  UseMethod("class_intervals")
}

#' @export
class_intervals.grd_classes <- function(x, ...) {
  attr(x, "class_intervals")
}

#' @export
class_intervals.grd_markov <- function(x, ...) {
  x$class_intervals
}

#' @export
class_intervals.grd_spatial_markov <- function(x, ...) {
  x$class_intervals
}

#' Spatial Lag Intervals Used By A Spatial Markov Object
#'
#' @param x A `grd_spatial_markov` object.
#'
#' @return A tibble of spatial-lag class intervals.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(1:4, each = 2),
#'   year = rep(2020:2021, times = 4),
#'   value = c(1, 2, 2, 3, 4, 3, 5, 6)
#' )
#' grid <- sf::st_sf(
#'   id = 1:4,
#'   geometry = sf::st_make_grid(
#'     sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2)),
#'     n = c(2, 2)
#'   )
#' ) |>
#'   dplyr::mutate(
#'     nb = sfdep::st_contiguity(geometry),
#'     wt = sfdep::st_weights(nb)
#'   )
#' spatial <- spatial_markov(panel, id, year, value, geometry = grid, k = 2)
#' lag_intervals(spatial)
#' @export
lag_intervals <- function(x) {
  UseMethod("lag_intervals")
}

#' @export
lag_intervals.grd_spatial_markov <- function(x) {
  x$lag_intervals
}

test_that("classify_dynamics creates ordered classes from long data", {
  dat <- tibble::tibble(
    id = rep(letters[1:4], each = 2),
    year = rep(c(2000, 2001), times = 4),
    value = c(1, 2, 2, 3, 3, 4, 4, 5)
  )

  out <- classify_dynamics(dat, id, year, value, k = 2)

  expect_s3_class(out, "grd_classes")
  expect_true(is.ordered(out$class))
  expect_equal(levels(out$class), c("Q1", "Q2"))
  expect_equal(nrow(class_intervals(out)), 2)
  expect_equal(class_intervals(out)$type, c("value", "value"))
})

test_that("markov_dynamics counts obvious transitions", {
  dat <- tibble::tibble(
    id = rep(c("a", "b", "c"), each = 2),
    year = rep(c(1, 2), times = 3),
    state = factor(c("low", "high", "low", "low", "high", "high"), levels = c("low", "high"), ordered = TRUE)
  )

  mk <- markov_dynamics(dat, id, year, state)

  expect_s3_class(mk, "grd_markov")
  expect_equal(unname(mk$counts), matrix(c(1, 1, 0, 1), nrow = 2, byrow = TRUE))
  expect_equal(unname(mk$probabilities), matrix(c(0.5, 0.5, 0, 1), nrow = 2, byrow = TRUE))
  expect_equal(sum(mk$counts), 3)
  expect_equal(mk$transitions$transition, c("low -> high", "low -> low", "high -> high"))
  expect_true("transition" %in% names(as.data.frame(mk)))
})

test_that("empty transition rows are explicit", {
  dat <- tibble::tibble(
    id = rep(c("a", "b"), each = 2),
    year = rep(c(1, 2), times = 2),
    state = factor(c("low", "low", "low", "low"), levels = c("low", "high"), ordered = TRUE)
  )

  mk <- markov_dynamics(dat, id, year, state)

  expect_equal(unname(mk$counts), matrix(c(2, 0, 0, 0), nrow = 2))
  expect_true(all(is.na(mk$probabilities[2, ])))
})

test_that("steady_state computes a simple stationary distribution", {
  P <- matrix(c(0.5, 0.5, 0.25, 0.75), nrow = 2, byrow = TRUE)
  ss <- steady_state(P)
  expect_equal(unname(ss), c(1 / 3, 2 / 3), tolerance = 1e-8)
})

.grd_col_name <- function(arg) {
  rlang::as_name(rlang::ensym(arg))
}

`%||%` <- function(x, y) {
  if (is.null(x)) y else x
}

.grd_is_sf <- function(x) {
  inherits(x, "sf")
}

.grd_drop_geometry <- function(x) {
  if (.grd_is_sf(x)) {
    sf::st_drop_geometry(x)
  } else {
    x
  }
}

.grd_state_labels <- function(k, labels = NULL) {
  if (is.null(labels)) {
    paste0("Q", seq_len(k))
  } else {
    labels
  }
}

.grd_quantile_breaks <- function(x, k = 5) {
  stats::quantile(x, probs = seq(0, 1, length.out = k + 1), na.rm = TRUE, type = 7)
}

.grd_break_table <- function(breaks, labels = NULL, type = "value") {
  breaks <- unique(as.numeric(breaks))
  labels <- .grd_state_labels(length(breaks) - 1, labels)
  tibble::tibble(
    class = factor(labels, levels = labels, ordered = TRUE),
    lower = breaks[-length(breaks)],
    upper = breaks[-1],
    type = type
  )
}

.grd_cut <- function(x, breaks, labels = NULL) {
  breaks <- unique(as.numeric(breaks))
  if (length(breaks) < 2) {
    stop("At least two unique break values are required.", call. = FALSE)
  }
  n_classes <- length(breaks) - 1
  labels <- .grd_state_labels(n_classes, labels)
  if (length(labels) != n_classes) {
    stop("`labels` must have one fewer value than `breaks`.", call. = FALSE)
  }
  cut(
    x,
    breaks = breaks,
    labels = labels,
    include.lowest = TRUE,
    ordered_result = TRUE
  )
}

.grd_ntile_factor <- function(x, k, labels = NULL) {
  labels <- .grd_state_labels(k, labels)
  out <- dplyr::ntile(x, k)
  factor(labels[out], levels = labels, ordered = TRUE)
}

.grd_prob_table <- function(counts) {
  probs <- prop.table(counts, margin = 1)
  probs[is.nan(probs)] <- NA_real_
  probs
}

.grd_complete_matrix_tibble <- function(transitions, state_levels) {
  transitions |>
    dplyr::count(.data$from_state, .data$to_state, name = "n") |>
    tidyr::complete(
      from_state = factor(state_levels, levels = state_levels, ordered = TRUE),
      to_state = factor(state_levels, levels = state_levels, ordered = TRUE),
      fill = list(n = 0L)
    ) |>
    dplyr::group_by(.data$from_state) |>
    dplyr::mutate(probability = if (sum(.data$n) > 0) .data$n / sum(.data$n) else NA_real_) |>
    dplyr::ungroup() |>
    dplyr::mutate(
      transition = paste(.data$from_state, "->", .data$to_state),
      .after = "to_state"
    )
}

.grd_stationary <- function(P, tol = 1e-12) {
  P <- as.matrix(P)
  if (nrow(P) != ncol(P)) {
    stop("Transition matrix must be square.", call. = FALSE)
  }
  if (anyNA(P)) {
    return(rep(NA_real_, nrow(P)))
  }
  eig <- eigen(t(P))
  idx <- which.min(Mod(eig$values - 1))
  v <- Re(eig$vectors[, idx])
  if (sum(v) < 0) v <- -v
  v[abs(v) < tol] <- 0
  v / sum(v)
}

#' Compute Rank Mobility
#'
#' Computes simple rank changes for map-ready exploratory analysis.
#'
#' @param data A long data frame or `sf` object.
#' @param id,time,value Columns identifying spatial unit, time, and value.
#' @param compare `"endpoint"` compares first and last observed periods per
#'   unit. `"adjacent"` compares adjacent periods.
#' @param descending If `TRUE`, larger values receive better ranks.
#'
#' @return A `grd_rank_mobility` data frame or `sf` object.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(letters[1:4], each = 3),
#'   year = rep(2020:2022, times = 4),
#'   value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
#' )
#'
#' rank_mobility(panel, id, year, value)
#' @export
rank_mobility <- function(data, id, time, value,
                          compare = c("endpoint", "adjacent"),
                          descending = TRUE) {
  compare <- match.arg(compare)
  id <- rlang::ensym(id)
  time <- rlang::ensym(time)
  value <- rlang::ensym(value)

  ranked <- data |>
    dplyr::group_by(!!time) |>
    dplyr::mutate(rank = if (descending) dplyr::min_rank(dplyr::desc(!!value)) else dplyr::min_rank(!!value)) |>
    dplyr::ungroup() |>
    dplyr::arrange(!!id, !!time)

  out <- if (compare == "endpoint") {
    ranked |>
      dplyr::group_by(!!id) |>
      dplyr::filter(!!time %in% range(!!time, na.rm = TRUE)) |>
      dplyr::mutate(
        start_time = dplyr::first(!!time),
        end_time = dplyr::last(!!time),
        start_rank = dplyr::first(.data$rank),
        end_rank = dplyr::last(.data$rank),
        rank_change = .data$start_rank - .data$end_rank,
        abs_rank_change = abs(.data$rank_change)
      ) |>
      dplyr::slice_tail(n = 1) |>
      dplyr::ungroup()
  } else {
    ranked |>
      dplyr::group_by(!!id) |>
      dplyr::mutate(
        to_time = dplyr::lead(!!time),
        to_rank = dplyr::lead(.data$rank),
        rank_change = .data$rank - .data$to_rank,
        abs_rank_change = abs(.data$rank_change)
      ) |>
      dplyr::ungroup() |>
      dplyr::filter(!is.na(.data$to_rank))
  }

  class(out) <- unique(c("grd_rank_mobility", class(out)))
  attr(out, "id") <- rlang::as_name(id)
  attr(out, "time") <- rlang::as_name(time)
  attr(out, "value") <- rlang::as_name(value)
  attr(out, "compare") <- compare
  out
}

#' @export
print.grd_rank_mobility <- function(x, ...) {
  cat("<grd_rank_mobility>\n")
  NextMethod()
}

#' @export
as.data.frame.grd_rank_mobility <- function(x, ...) {
  class(x) <- setdiff(class(x), "grd_rank_mobility")
  as.data.frame(x, ...)
}

#' US State Per-Capita Personal Income, 1929-2009
#'
#' Per-capita personal income for the 48 contiguous US states in nominal
#' dollars. Mirrors the canonical `usjoin` panel used in PySAL `giddy` and the
#' spatial Markov literature, so examples and validation in this package are
#' directly comparable to that reference work.
#'
#' @format A tibble with 3,888 rows (48 states x 81 years) and 4 columns:
#' \describe{
#'   \item{name}{State name.}
#'   \item{state_fips}{State FIPS code, integer.}
#'   \item{year}{Year, integer 1929 to 2009.}
#'   \item{income}{Nominal per-capita personal income, integer USD.}
#' }
#'
#' @source PySAL `libpysal` `examples/us_income/usjoin.csv`. The original series
#'   is constructed from US Bureau of Economic Analysis state personal income
#'   tables. See <https://github.com/pysal/libpysal>.
#'
#' @references
#' Rey, S. J. (2001). Spatial empirics for economic growth and convergence.
#' *Geographical Analysis*, 33(3), 195-214.
#'
#' Rey, S. J., Kang, W., & Wolf, L. (2016). The properties of tests for spatial
#' effects in discrete Markov chain models of regional income distribution
#' dynamics. *Journal of Geographical Systems*, 18(4), 377-398.
#'
#' @examples
#' data(usjoin)
#' head(usjoin)
#' range(usjoin$year)
"usjoin"

#' griddy: Geospatial Distribution Dynamics
#'
#' Tidy helpers for exploratory geospatial distribution dynamics.
#'
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom rlang .data
## usethis namespace: end
NULL

#' Estimate Spatial Markov Transition Dynamics
#'
#' Estimates transition matrices conditioned on the class of each unit's spatial
#' lag at the start of the transition period.
#'
#' @param data A long data frame or `sf` object.
#' @param id,time,value Columns identifying spatial unit, time, and value.
#' @param geometry An `sf` tibble with one row per spatial unit (a single time
#'   slice's geography), carrying `nb` and `wt` list-columns produced by
#'   `sfdep::st_contiguity()` and `sfdep::st_weights()`. The preferred input.
#' @param listw A row-standardized `spdep` `listw` object. Accepted for
#'   compatibility with prior workflows; use `geometry` for new code.
#' @param nb A `spdep` neighbor list, used only when both `geometry` and
#'   `listw` are `NULL`. Converted with `spdep::nb2listw(style = "W")`.
#' @param k Number of value classes.
#' @param lag_k Number of spatial-lag classes.
#' @param class_method Value classification method passed to [classify_dynamics()].
#' @param zero.policy Passed to `spdep` lag/weight helpers.
#'
#' @return A `grd_spatial_markov` object.
#'
#' @examplesIf identical(Sys.getenv("IN_PKGDOWN"), "true")
#' panel <- data.frame(
#'   id = rep(1:4, each = 3),
#'   year = rep(2020:2022, times = 4),
#'   value = c(1, 2, 3, 2, 3, 4, 4, 3, 5, 5, 6, 7)
#' )
#'
#' grid <- sf::st_sf(
#'   id = 1:4,
#'   geometry = sf::st_make_grid(
#'     sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2)),
#'     n = c(2, 2)
#'   )
#' ) |>
#'   dplyr::mutate(
#'     nb = sfdep::st_contiguity(geometry),
#'     wt = sfdep::st_weights(nb)
#'   )
#'
#' spatial <- spatial_markov(panel, id, year, value, geometry = grid, k = 2)
#'
#' spatial
#' lag_intervals(spatial)
#' transition_matrix(spatial, "count", lag_class = "Q1")
#' @export
spatial_markov <- function(data, id, time, value,
                           geometry = NULL,
                           listw = NULL,
                           nb = NULL,
                           k = 5,
                           lag_k = k,
                           class_method = c("pooled_quantile", "time_quantile", "fixed"),
                           zero.policy = TRUE) {
  class_method <- match.arg(class_method)
  id <- rlang::ensym(id)
  time <- rlang::ensym(time)
  value <- rlang::ensym(value)

  if (!is.null(geometry)) {
    if (!"nb" %in% names(geometry)) {
      stop(
        "`geometry` must carry an `nb` list-column. ",
        "Build one with `sfdep::st_contiguity()`.",
        call. = FALSE
      )
    }
    nb_obj <- geometry$nb
    if ("wt" %in% names(geometry)) {
      listw <- spdep::nb2listw(
        nb_obj,
        glist = geometry$wt,
        style = "W",
        zero.policy = zero.policy
      )
    } else {
      listw <- spdep::nb2listw(nb_obj, style = "W", zero.policy = zero.policy)
    }
  } else if (is.null(listw)) {
    if (is.null(nb)) {
      stop("Provide `geometry`, `listw`, or `nb`.", call. = FALSE)
    }
    listw <- spdep::nb2listw(nb, style = "W", zero.policy = zero.policy)
  }

  flat <- .grd_drop_geometry(data)
  id_name <- rlang::as_name(id)
  time_name <- rlang::as_name(time)
  value_name <- rlang::as_name(value)

  id_order <- flat |>
    dplyr::arrange(!!time, !!id) |>
    dplyr::filter(!!time == min(!!time, na.rm = TRUE)) |>
    dplyr::pull(!!id)

  n_units <- length(id_order)
  if (length(listw$neighbours) != n_units) {
    stop("The weights object length must match the number of spatial units.", call. = FALSE)
  }

  value_classes <- classify_dynamics(
    flat,
    id = !!id,
    time = !!time,
    value = !!value,
    method = class_method,
    k = k
  )
  value_intervals <- attr(value_classes, "class_intervals")
  value_classes <- value_classes |>
    .grd_drop_geometry() |>
    dplyr::select(!!id, !!time, "class") |>
    dplyr::rename(value_class = "class")

  lag_panel <- flat |>
    dplyr::select(!!id, !!time, !!value) |>
    dplyr::group_by(!!time) |>
    dplyr::group_modify(function(.x, .y) {
      arranged <- .x |>
        dplyr::arrange(match(.data[[id_name]], id_order))
      if (!identical(as.character(arranged[[id_name]]), as.character(id_order))) {
        stop("Each time period must contain the same IDs as the weights object order.", call. = FALSE)
      }
      arranged$spatial_lag <- spdep::lag.listw(listw, arranged[[value_name]], zero.policy = zero.policy)
      arranged
    }) |>
    dplyr::ungroup()

  lag_breaks <- .grd_quantile_breaks(lag_panel$spatial_lag, lag_k)
  lag_intervals <- .grd_break_table(lag_breaks, type = "spatial_lag")

  lag_panel <- lag_panel |>
    dplyr::mutate(lag_class = .grd_cut(.data$spatial_lag, lag_breaks)) |>
    dplyr::left_join(value_classes, by = stats::setNames(c(id_name, time_name), c(id_name, time_name)))

  state_levels <- levels(value_classes$value_class)
  lag_levels <- levels(lag_panel$lag_class)

  transitions <- lag_panel |>
    dplyr::arrange(!!id, !!time) |>
    dplyr::group_by(!!id) |>
    dplyr::mutate(
      to_state = dplyr::lead(.data$value_class),
      to_time = dplyr::lead(!!time)
    ) |>
    dplyr::ungroup() |>
    dplyr::filter(!is.na(.data$to_state)) |>
    dplyr::transmute(
      id = !!id,
      from_time = !!time,
      to_time = .data$to_time,
      lag_class = factor(as.character(.data$lag_class), levels = lag_levels, ordered = TRUE),
      from_state = factor(as.character(.data$value_class), levels = state_levels, ordered = TRUE),
      to_state = factor(as.character(.data$to_state), levels = state_levels, ordered = TRUE),
      spatial_lag = .data$spatial_lag
    ) |>
    dplyr::mutate(
      transition = paste(.data$from_state, "->", .data$to_state),
      .after = "to_state"
    )

  matrix_data <- transitions |>
    dplyr::count(.data$lag_class, .data$from_state, .data$to_state, name = "n") |>
    tidyr::complete(
      lag_class = factor(lag_levels, levels = lag_levels, ordered = TRUE),
      from_state = factor(state_levels, levels = state_levels, ordered = TRUE),
      to_state = factor(state_levels, levels = state_levels, ordered = TRUE),
      fill = list(n = 0L)
    ) |>
    dplyr::group_by(.data$lag_class, .data$from_state) |>
    dplyr::mutate(probability = if (sum(.data$n) > 0) .data$n / sum(.data$n) else NA_real_) |>
    dplyr::ungroup() |>
    dplyr::mutate(
      transition = paste(.data$from_state, "->", .data$to_state),
      .after = "to_state"
    ) |>
    dplyr::left_join(
      lag_intervals |>
        dplyr::rename(lag_class = "class", lag_lower = "lower", lag_upper = "upper") |>
        dplyr::select("lag_class", "lag_lower", "lag_upper"),
      by = "lag_class"
    )

  out <- list(
    transitions = transitions,
    matrix = matrix_data,
    lag_panel = lag_panel,
    states = state_levels,
    lag_states = lag_levels,
    lag_breaks = lag_breaks,
    class_intervals = value_intervals,
    lag_intervals = lag_intervals,
    id = id_name,
    time = time_name,
    value = value_name,
    listw = listw
  )
  class(out) <- "grd_spatial_markov"
  out
}

#' @export
print.grd_spatial_markov <- function(x, ...) {
  cat("<grd_spatial_markov>\n")
  cat(length(x$states), "states,", length(x$lag_states), "lag states,", nrow(x$transitions), "observed transitions\n")
  print(x$matrix)
  invisible(x)
}

#' @export
as.data.frame.grd_spatial_markov <- function(x, ...) {
  as.data.frame(x$matrix, ...)
}

---
title: "Geospatial Distribution Dynamics With griddy"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Geospatial Distribution Dynamics With griddy}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, message=FALSE, warning=FALSE}
library(griddy)
library(dplyr)
library(ggplot2)
library(sf)
library(sfdep)
library(spData)
library(tidyr)
```

```{r map-theme, include=FALSE}
map_theme <- function() {
  theme_void(base_size = 11) +
    theme(
      legend.position = "bottom",
      plot.title.position = "plot",
      strip.text = element_text(face = "bold")
    )
}
```

`griddy` works on long spatial panels: one row per place per period, keyed by
explicit `id`, `time`, and `value` columns. The bundled `usjoin` panel contains
48 contiguous US states, per-capita personal income, and annual observations
from 1929 to 2009. It mirrors the canonical PySAL `giddy` example used in
Rey (2001).

```{r}
data(usjoin)

usjoin
```

## Classification

Pooled quantile classification cuts the entire panel into five bins and reuses
the same breakpoints for every year. That keeps class meanings comparable
across periods, which matters because year-specific quantiles can hide absolute
movement. A state in Q5 in 2009 is being compared to the same income scale as
a state in Q5 in 1929, not merely to its contemporaries.

```{r}
classes <- classify_dynamics(usjoin, name, year, income, k = 5)

class_intervals(classes)
```

Nominal income growth means the endpoints are nearly homogeneous under pooled
breaks: the early panel sits near the bottom of the pooled distribution and the
late panel sits near the top. A mid-panel year is a better mapped diagnostic
because it shows both the absolute class scale and the cross-state ordering.

```{r}
states <- us_states |>
  st_drop_geometry() |>
  filter(NAME %in% usjoin$name) |>
  pull(NAME)

geom <- us_states |>
  filter(NAME %in% usjoin$name) |>
  arrange(NAME)
```

```{r, fig.alt="Map of contiguous US states showing pooled income classes in 1955 using the same five class intervals estimated from the full 1929 to 2009 panel.", fig.width=6.8, fig.height=4}
class_map <- classes |>
  filter(year == 1955) |>
  left_join(geom |> select(NAME, geometry), by = c("name" = "NAME")) |>
  st_as_sf()

ggplot(class_map) +
  geom_sf(aes(fill = class), color = "white", linewidth = 0.15) +
  scale_fill_viridis_d(option = "C", direction = -1, name = "Income class") +
  coord_sf(datum = NA) +
  labs(
    title = "Pooled classes remain map-ready",
    subtitle = "1955 uses breaks estimated from the full 1929-2009 panel"
  ) +
  map_theme()
```

## Classic Markov

The classic Markov matrix estimates movement between income classes from one
year to the next, pooling transitions across all states and adjacent years.
The `print()` method gives the transition probabilities; `transition_matrix()`
can return either probabilities or raw counts for downstream work.

```{r, fig.alt="Heatmap of classic Markov transition probabilities for US state per-capita income, 1929 to 2009."}
classic <- markov_dynamics(classes, name, year, class)

classic
transition_matrix(classic, "count")
steady_state(classic)

plot_transition_matrix(classic)
```

The diagonal dominance of the matrix — most states stay in their starting
quintile from one year to the next — is the headline finding from this panel
and recovers the result Rey (2001) reports. The stationary distribution is a
mechanical summary of this empirical matrix, not a forecast. Here Q5 is
absorbing in the observed one-year transitions, so the long-run vector puts all
mass in Q5.

## Spatial Markov

Spatial Markov conditions transition probabilities on the starting class of
each unit's spatial lag. The expectation is that being surrounded by rich (or
poor) neighbors shifts the transition odds beyond what the unit's own class
would predict.

```{r}
geom <- us_states |>
  filter(NAME %in% usjoin$name) |>
  arrange(NAME) |>
  mutate(
    nb = st_contiguity(geometry),
    wt = st_weights(nb)
  )

panel <- usjoin |>
  filter(name %in% states) |>
  arrange(name, year)
```

`geometry` is the preferred way to pass spatial weights: an `sf` tibble with
one row per unit and `nb` / `wt` list-columns produced by `sfdep`. Carrying
weights as columns on the geography keeps the spatial frame, the weights, and
their row-standardization choice in one tidy object.

```{r, fig.alt="Faceted heatmaps of spatial Markov transition probabilities for US state per-capita income, by spatial-lag quintile."}
spatial <- spatial_markov(panel, name, year, income, geometry = geom, k = 5)

lag_intervals(spatial)
transition_matrix(spatial, "probability", lag_class = "Q1")

plot_spatial_markov(spatial)
```

Reading the facets left to right: states whose neighbors sit in the lowest
income lag class (Q1) are much stickier at the bottom of the distribution than
the pooled transition matrix would suggest. The spatial context conditioning is
what spatial Markov is built to surface. Grey rows indicate lag-class/state
combinations with no observed transitions, so their probabilities are undefined
rather than zero.

The lag classes themselves are map-ready. The map below uses 1994 because the
spatial-lag distribution is split across two classes in that year; at the start
of the panel, all states have Q1 spatial lags under pooled nominal-income
breaks. Each state's class below is based on the weighted income of its
neighbors, not on its own income.

```{r, fig.alt="Map of contiguous US states showing each state's 1994 spatial-lag income class based on neighboring states.", fig.width=6.8, fig.height=4}
lag_map <- spatial$transitions |>
  filter(from_time == 1994) |>
  distinct(id, lag_class, spatial_lag) |>
  left_join(geom |> select(NAME, geometry), by = c("id" = "NAME")) |>
  st_as_sf()

ggplot(lag_map) +
  geom_sf(aes(fill = lag_class), color = "white", linewidth = 0.15) +
  scale_fill_viridis_d(option = "C", direction = -1, name = "Spatial-lag class") +
  coord_sf(datum = NA) +
  labs(
    title = "Spatial Markov conditions on neighboring-state income",
    subtitle = "Spatial-lag class in 1994"
  ) +
  map_theme()
```

## Rank Mobility

Rank mobility is intentionally simple in v0.1: an endpoint comparison of unit
ranks between the first and last observed periods. It is descriptive and
map-ready, not the full PySAL rank-method suite. Positive values mean a state
moved up the income ranking; negative values mean it moved down.

```{r, fig.alt="Map of endpoint rank mobility for US states, 1929 to 2009, with positive values indicating upward rank movement."}
mobility_panel <- usjoin |>
  filter(name %in% states) |>
  left_join(geom |> select(NAME), by = c("name" = "NAME")) |>
  st_as_sf()

mobility <- rank_mobility(mobility_panel, name, year, income)

mobility |>
  st_drop_geometry() |>
  arrange(desc(abs_rank_change)) |>
  select(name, start_rank, end_rank, rank_change) |>
  head()

plot_rank_mobility(mobility) +
  coord_sf(datum = NA) +
  labs(
    title = "Endpoint rank mobility, 1929 to 2009",
    subtitle = "Positive values indicate upward movement in the state income ranking"
  ) +
  map_theme()
```

For shorter-run churn, use adjacent-period comparisons. The same result schema
is returned, but each row describes movement from one observed period to the
next.

```{r}
adjacent_mobility <- rank_mobility(panel, name, year, income, compare = "adjacent")

adjacent_mobility |>
  arrange(desc(abs_rank_change)) |>
  select(name, year, to_time, rank_change) |>
  head()
```

## Output schemas

Each result object exposes a tidy `transitions` table and a long `matrix` table
keyed by class labels. Inspecting them directly is the simplest way to learn
the schema.

```{r}
classic$transitions |> head()
spatial$transitions |> select(id, from_time, to_time, lag_class, transition, spatial_lag) |> head()
```

For programmatic access, every griddy object also has `as.data.frame()` and the
S3 generics `transition_matrix()`, `class_intervals()`, and (for spatial
objects) `lag_intervals()`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatial-markov.R
\name{spatial_markov}
\alias{spatial_markov}
\title{Estimate Spatial Markov Transition Dynamics}
\usage{
spatial_markov(
  data,
  id,
  time,
  value,
  geometry = NULL,
  listw = NULL,
  nb = NULL,
  k = 5,
  lag_k = k,
  class_method = c("pooled_quantile", "time_quantile", "fixed"),
  zero.policy = TRUE
)
}
\arguments{
\item{data}{A long data frame or \code{sf} object.}

\item{id, time, value}{Columns identifying spatial unit, time, and value.}

\item{geometry}{An \code{sf} tibble with one row per spatial unit (a single time
slice's geography), carrying \code{nb} and \code{wt} list-columns produced by
\code{sfdep::st_contiguity()} and \code{sfdep::st_weights()}. The preferred input.}

\item{listw}{A row-standardized \code{spdep} \code{listw} object. Accepted for
compatibility with prior workflows; use \code{geometry} for new code.}

\item{nb}{A \code{spdep} neighbor list, used only when both \code{geometry} and
\code{listw} are \code{NULL}. Converted with \code{spdep::nb2listw(style = "W")}.}

\item{k}{Number of value classes.}

\item{lag_k}{Number of spatial-lag classes.}

\item{class_method}{Value classification method passed to \code{\link[=classify_dynamics]{classify_dynamics()}}.}

\item{zero.policy}{Passed to \code{spdep} lag/weight helpers.}
}
\value{
A \code{grd_spatial_markov} object.
}
\description{
Estimates transition matrices conditioned on the class of each unit's spatial
lag at the start of the transition period.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(1:4, each = 3),
  year = rep(2020:2022, times = 4),
  value = c(1, 2, 3, 2, 3, 4, 4, 3, 5, 5, 6, 7)
)

grid <- sf::st_sf(
  id = 1:4,
  geometry = sf::st_make_grid(
    sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2)),
    n = c(2, 2)
  )
) |>
  dplyr::mutate(
    nb = sfdep::st_contiguity(geometry),
    wt = sfdep::st_weights(nb)
  )

spatial <- spatial_markov(panel, id, year, value, geometry = grid, k = 2)

spatial
lag_intervals(spatial)
transition_matrix(spatial, "count", lag_class = "Q1")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_transition_matrix}
\alias{plot_transition_matrix}
\title{Plot A Transition Matrix}
\usage{
plot_transition_matrix(x)
}
\arguments{
\item{x}{A \code{grd_markov} object.}
}
\value{
A \code{ggplot} object.
}
\description{
Plot A Transition Matrix
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(id = rep(1:4, each = 2), year = rep(2020:2021, 4), value = 1:8)
markov <- panel |>
  classify_dynamics(id, year, value, k = 2) |>
  markov_dynamics(id, year, class)

plot_transition_matrix(markov)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markov.R
\name{lag_intervals}
\alias{lag_intervals}
\title{Spatial Lag Intervals Used By A Spatial Markov Object}
\usage{
lag_intervals(x)
}
\arguments{
\item{x}{A \code{grd_spatial_markov} object.}
}
\value{
A tibble of spatial-lag class intervals.
}
\description{
Spatial Lag Intervals Used By A Spatial Markov Object
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(1:4, each = 2),
  year = rep(2020:2021, times = 4),
  value = c(1, 2, 2, 3, 4, 3, 5, 6)
)
grid <- sf::st_sf(
  id = 1:4,
  geometry = sf::st_make_grid(
    sf::st_bbox(c(xmin = 0, ymin = 0, xmax = 2, ymax = 2)),
    n = c(2, 2)
  )
) |>
  dplyr::mutate(
    nb = sfdep::st_contiguity(geometry),
    wt = sfdep::st_weights(nb)
  )
spatial <- spatial_markov(panel, id, year, value, geometry = grid, k = 2)
lag_intervals(spatial)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rank-mobility.R
\name{rank_mobility}
\alias{rank_mobility}
\title{Compute Rank Mobility}
\usage{
rank_mobility(
  data,
  id,
  time,
  value,
  compare = c("endpoint", "adjacent"),
  descending = TRUE
)
}
\arguments{
\item{data}{A long data frame or \code{sf} object.}

\item{id, time, value}{Columns identifying spatial unit, time, and value.}

\item{compare}{\code{"endpoint"} compares first and last observed periods per
unit. \code{"adjacent"} compares adjacent periods.}

\item{descending}{If \code{TRUE}, larger values receive better ranks.}
}
\value{
A \code{grd_rank_mobility} data frame or \code{sf} object.
}
\description{
Computes simple rank changes for map-ready exploratory analysis.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(letters[1:4], each = 3),
  year = rep(2020:2022, times = 4),
  value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
)

rank_mobility(panel, id, year, value)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markov.R
\name{markov_dynamics}
\alias{markov_dynamics}
\title{Estimate Classic Markov Transition Dynamics}
\usage{
markov_dynamics(classes, id, time, state = class)
}
\arguments{
\item{classes}{A classified data frame from \code{\link[=classify_dynamics]{classify_dynamics()}} or any data
frame with ID, time, and state columns.}

\item{id, time, state}{Columns identifying spatial unit, time, and state.}
}
\value{
A \code{grd_markov} object.
}
\description{
Estimates transition counts and probabilities between adjacent periods for a
classified long panel.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(letters[1:4], each = 3),
  year = rep(2020:2022, times = 4),
  value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
)

classes <- classify_dynamics(panel, id, year, value, k = 3)
markov <- markov_dynamics(classes, id, year, class)

markov
transition_matrix(markov)
steady_state(markov)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markov.R
\name{transition_matrix}
\alias{transition_matrix}
\title{Extract A Transition Matrix}
\usage{
transition_matrix(x, type = c("probability", "count"), lag_class = NULL)
}
\arguments{
\item{x}{A \code{grd_markov} or \code{grd_spatial_markov} object.}

\item{type}{\code{"probability"} or \code{"count"}.}

\item{lag_class}{Optional lag class for \code{grd_spatial_markov}.}
}
\value{
A matrix.
}
\description{
Extract A Transition Matrix
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(letters[1:4], each = 3),
  year = rep(2020:2022, times = 4),
  value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
)
markov <- panel |>
  classify_dynamics(id, year, value, k = 3) |>
  markov_dynamics(id, year, class)

transition_matrix(markov, "count")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markov.R
\name{class_intervals}
\alias{class_intervals}
\title{Class Intervals Used By A griddy Object}
\usage{
class_intervals(x, ...)
}
\arguments{
\item{x}{A griddy result object.}

\item{...}{Reserved for future methods.}
}
\value{
A tibble of class intervals when available.
}
\description{
Class Intervals Used By A griddy Object
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(id = rep(1:3, each = 2), year = rep(2020:2021, 3), value = 1:6)
classes <- classify_dynamics(panel, id, year, value, k = 3)
class_intervals(classes)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classify.R
\name{classify_dynamics}
\alias{classify_dynamics}
\title{Classify Longitudinal Values For Distribution Dynamics}
\usage{
classify_dynamics(
  data,
  id,
  time,
  value,
  method = c("pooled_quantile", "time_quantile", "fixed", "existing"),
  k = 5,
  breaks = NULL,
  labels = NULL
)
}
\arguments{
\item{data}{A data frame or \code{sf} object in long format.}

\item{id, time, value}{Columns identifying spatial unit, time, and value.}

\item{method}{Classification method. \code{"pooled_quantile"} uses one set of
quantile breaks across all periods. \code{"time_quantile"} uses period-specific
ranks. \code{"fixed"} uses user-supplied \code{breaks}. \code{"existing"} treats \code{value}
as an already classified state.}

\item{k}{Number of quantile classes.}

\item{breaks}{Breaks for \code{method = "fixed"}.}

\item{labels}{Optional class labels.}
}
\value{
A data frame or \code{sf} object with a \code{class} column and class
\code{grd_classes}.
}
\description{
Converts numeric values in a long panel into ordered classes suitable for
transition analysis.
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
panel <- data.frame(
  id = rep(letters[1:4], each = 3),
  year = rep(2020:2022, times = 4),
  value = c(8, 9, 11, 10, 12, 13, 15, 14, 16, 20, 22, 25)
)

classes <- classify_dynamics(panel, id, year, value, k = 3)
classes
class_intervals(classes)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{plot_rank_mobility}
\alias{plot_rank_mobility}
\title{Plot Rank Mobility}
\usage{
plot_rank_mobility(x)
}
\arguments{
\item{x}{A \code{grd_rank_mobility} object returned by \code{\link[=rank_mobility]{rank_mobility()}}.}
}
\value{
A \code{ggplot} object.
}
\description{
Plot Rank Mobility
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markov.R
\name{steady_state}
\alias{steady_state}
\title{Estimate Stationary Distribution}
\usage{
steady_state(x)
}
\arguments{
\item{x}{A \code{grd_markov} object or transition probability matrix.}
}
\value{
A numeric vector.
}
\description{
Estimate Stationary Distribution
}
\examples{
\dontshow{if (identical(Sys.getenv("IN_PKGDOWN"), "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
prob <- matrix(c(0.8, 0.2, 0.3, 0.7), nrow = 2, byrow = TRUE)
steady_state(prob)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{usjoin}
\alias{usjoin}
\title{US State Per-Capita Personal Income, 1929-2009}
\format{
A tibble with 3,888 rows (48 states x 81 years) and 4 columns:
\describe{
\item{name}{State name.}
\item{state_fips}{State FIPS code, integer.}
\item{year}{Year, integer 1929 to 2009.}
\item{income}{Nominal per-capita personal income, integer USD.}
}
}
\source{
PySAL \code{libpysal} \code{examples/us_income/usjoin.csv}. The original series
is constructed from US Bureau of Economic Analysis state personal income
tables. See \url{https://github.com/pysal/libpysal}.
}
\usage{
usjoin
}
\description{
Per-capita personal income for the 48 contiguous US states in nominal
dollars. Mirrors the canonical \code{usjoin} panel used in PySAL \code{giddy} and the
spatial Markov literature, so examples and validation in this package are
directly comparable to that reference work.
}
\examples{
data(usjoin)
head(usjoin)
range(usjoin$year)
}
\references{
Rey, S. J. (2001). Spatial empirics for economic growth and convergence.
\emph{Geographical Analysis}, 33(3), 195-214.

Rey, S. J., Kang, W., & Wolf, L. (2016). The properties of tests for spatial
effects in discrete Markov chain models of regional income distribution
dynamics. \emph{Journal of Geographical Systems}, 18(4), 377-398.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/griddy-package.R
\docType{package}
\name{griddy-package}
\alias{griddy}
\alias{griddy-package}
\title{griddy: Geospatial Distribution Dynamics}
\description{
Tidy helpers for exploratory geospatial distribution dynamics.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/dshkol/griddy}
  \item \url{https://dshkol.github.io/griddy/}
  \item Report bugs at \url{https://github.com/dshkol/griddy/issues}
}

}
\author{
\strong{Maintainer}: Dmitry Shkolnik \email{shkolnikd@gmail.com}

}
\keyword{internal}
